pub(crate) const XML_DECLARATION: &str =
    r#"<?xml version="1.0" encoding="UTF-8" standalone="yes"?>"#;

pub(crate) const XML_WORKSHEET: &str =
    r#"http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet"#;
